@@ -68,7 +68,24 @@ loadTest("reduce_builtin.js");
 loadTest("reduce_false.js");
 loadTest("reduce_false_temp.js");
 loadTest("replication.js");
-loadTest("replicator_db.js");
+loadScript("script/replicator_db_inc.js");
+loadTest("replicator_db_bad_rep_id.js");
+loadTest("replicator_db_by_doc_id.js");
+loadTest("replicator_db_compact_rep_db.js");
+loadTest("replicator_db_continuous.js");
+loadTest("replicator_db_credential_delegation.js");
+loadTest("replicator_db_field_validation.js");
+loadTest("replicator_db_filtered.js");
+loadTest("replicator_db_identical.js");
+loadTest("replicator_db_identical_continuous.js");
+loadTest("replicator_db_invalid_filter.js");
+loadTest("replicator_db_simple.js");
+loadTest("replicator_db_successive.js");
+loadTest("replicator_db_survives.js");
+loadTest("replicator_db_swap_rep_db.js");
+loadTest("replicator_db_update_security.js");
+loadTest("replicator_db_user_ctx.js");
+loadTest("replicator_db_write_auth.js");
 loadTest("replicator_db_security.js");
 loadTest("rev_stemming.js");
 loadTest("rewrite.js");
